#
# Copyright (c) 2021-2022 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Wed Dec 15 2021
#
import sys
import traceback

# --- iscan ----------------------------------------------------------------
#
if 'readline' in sys.modules:
  import readline

  def iscan(prompt):
    """Read value interactively."""
    hlen = readline.get_current_history_length()

    arg = input(prompt + ' > ')

    try: readline.remove_history_item(hlen)
    except ValueError: pass

    return arg

else:

  def iscan(prompt):
    """Read value interactively."""
    return input(prompt)


# --- iscanarg -------------------------------------------------------------
#
def iscanarg(kind, prompt, defval, length = None, values = None):
  """Read a scalar argument interactively."""

  if defval:
    prompt += ' (%s)' % str(defval)

  while True:
    arg = iscan(prompt)

    # user might want default value or interrupt
    if not arg: arg = str(defval)
    if arg == '...': return ...

    # check input validity
    try:
      if kind in ('char', 'string', 'enum'):
        arg = arg.format_map(globals())
      else:
        arg = eval(arg, globals(), {})

      if kind == 'boolean':
        arg = bool(arg)
      elif kind == 'octet':
        arg = int(arg)
        if not -128 <= arg <= 127:
          raise OverflowError('expecting octet')
      elif kind == 'short':
        arg = int(arg)
        if not -32768 <= arg <= 32767:
          raise OverflowError('expecting short')
      elif kind == 'unsigned short':
        arg = int(arg)
        if not 0 <= arg <= 65536:
          raise OverflowError('expecting unsigned short')
      elif kind == 'long':
        arg = int(arg)
        if not -2**31 <= arg <= 2**31-1:
          raise OverflowError('expecting long')
      elif kind == 'unsigned long':
        arg = int(arg)
        if not 0 <= arg <= 2**32-1:
          raise OverflowError('expecting unsigned long')
      elif kind == 'long long':
        arg = int(arg)
        if not -2**63 <= arg <= 2**63-1:
          raise OverflowError('expecting long')
      elif kind == 'unsigned long long':
        arg = int(arg)
        if not 0 <= arg <= 2**64-1:
          raise OverflowError('expecting unsigned lnog long')
      elif kind in ('float', 'double'):
        arg = float(arg)
      elif kind == 'char':
        arg = str(arg)
        if len(arg) != 1:
          raise OverflowError('expecting char')
      elif kind == 'string':
        arg = str(arg)
        if length is not None:
          if len(arg) > length:
            raise OverflowError('string too long')
      elif kind == 'enum':
        arg = str(arg).strip()
        if arg not in values:
          raise KeyError(', '.join(values))
      else:
        raise TypeError(kind)

    except Exception as e:
      traceback.print_exception(type(e), e, None)

    else: return arg


# --- iscandict ----------------------------------------------------------
#
def iscandict(indent, name, fmt, meta):
  """Read a dictionary of arguments interactively."""

  kind = fmt['kind']
  items = meta.get('items', {})
  defval = meta.get('default', '')

  prompt = indent + kind + ' ' + name
  if 'length' in fmt: prompt += '[%s]' % fmt['length']
  if 'doc' in meta: prompt += ': ' + meta['doc']

  if kind == 'struct':
    print(prompt)

    arg = {}
    for p, f in fmt['type'].items():
      m = items.get(p, {})
      v = iscandict(indent + "| ", name + '.' + p, f, m)
      if v is ...: break

      arg[p] = v

  elif kind == 'union':
    print(prompt)

    arg = {}
    while True:
      # get discriminant
      v = iscandict(indent + "| ", name + '.d', fmt['discr'], {})
      if v is ...: break
      arg['d'] = v

      # find case
      p = next((p for p, c in fmt['cases'].items() if c == v), None)
      if p is None: continue

      f = fmt['type'].get(p)
      if f is None: continue

      m = items.get(p, {})
      v = iscandict(indent + "| ", name + '.u.' + p, f, m)
      if v is ...: break

      arg[u] = {p: v}
      break

  elif kind in ('array', 'sequence'):
    print(prompt)

    arg = []
    length = fmt.get('length')
    if length is None:
      length = iscanarg('unsigned long', prompt + ' length', 0)
      if length is ...: return arg

    for i in range(0, length):
      m = items.get(str(i), {})
      v = iscandict(indent + "| ", name + '[%s]' % i, fmt['type'], m)
      if v is ...: break

      arg += [v]

  elif kind == 'optional':
    print(prompt)

    arg = []
    present = iscanarg('boolean', indent + '| present', 1)
    if present is ...: return arg

    if present:
      m = {'items': items, 'default': defval}
      v = iscandict(indent + "| ", name, fmt['type'], m)
      if v is ...: return arg

      arg = v

  else:
    length = fmt.get('length')
    values = fmt.get('values')

    arg = iscanarg(kind, prompt, defval, length = length, values = values)

  return arg


# --- iscanargs ----------------------------------------------------------
#
def iscanargs(values, meta):
  """Read service arguments interactively."""
  arg = {}
  for p, f in values.items():
    m = meta.get(p, {})
    v = iscandict("", p, f, m)
    if v is ...: break

    arg[p] = v

  return arg


# --- args2json ------------------------------------------------------------
#
def args2json(name, fmt, args, kwargs, defarg):
  """Convert service arguments to a json string, merging flat list and
  defaults.
  """
  kind = fmt['kind']

  if kind in ('struct'):
    # check arguments
    if not isinstance(kwargs, dict):
      if kwargs is not ...:
        raise TypeError('%s expects a dictionary, not %s' %
                        ('.'.join(name), str(type(kwargs))))
      kwargs = {}

    if defarg is ...:
      defarg = {}

    # recurse
    json = []
    for p, f in fmt['type'].items():
      v = '"' + p + '":' + args2json(
        name + (p,), f, args, kwargs.pop(p, ...), defarg.get(p, ...))
      json.append(v)

    if kwargs:
      p, f = kwargs.popitem()
      raise TypeError('unknown argument %s.%s' % ('.'.join(name), p))

    return '{' + ','.join(json) + '}'

  elif kind in ('union'):
    if not isinstance(kwargs, dict):
      if kwargs is not ...:
        raise TypeError('%s expects a dictionary, not %s' %
                        ('.'.join(name), str(type(kwargs))))
      kwargs = {}

    if defarg is ...:
      defarg = {}

    # recurse
    d = args2json(name + ('d',), fmt['discr'],
                  args, kwargs.pop('d', ...), defarg.get('d', ...))
    u = kwargs.pop('u', {})

    json = []
    for p, f in fmt['type'].items():
      v = '"' + p + '":' + args2json(
        name + (p,), f, args, u.pop(p, ...), defarg.get(p, ...))
      json.append(v)

    return '{"d":' + d + ',"u":{' ','.join(json) + '}}'

  elif kind in ('array', 'sequence'):
    etype = fmt['type']

    # fetch value
    v = kwargs
    if v is ...:
      if len(args): v = args.pop(0)
      elif defarg is not ...: v = defarg
      else: raise TypeError('missing value for %s' % '.'.join(name))

    if not isinstance(v, (dict, list, tuple)):
      raise TypeError('%s expects a list or dictionary, not %s' %
                      ('.'.join(name), str(type(v))))

    # check arguments
    l = fmt.get('length', len(v))
    if len(v) > l:
      raise IndexError('%s too long (%d max)' % ('.'.join(name), l))

    # recurse
    if isinstance(v, (list, tuple)):
      if kind in ('array') and len(v) < l:
        raise IndexError('%s must have %d elements' % ('.'.join(name), l))
      json = []
      for i in range(0, l):
        e = args2json(name + ('[%d]' % i,), etype, [], v[i], ...)
        json.append(e)

      return '[' + ','.join(json) + ']'

    else: # dict
      json = []
      for i, x in v.items():
        if not i.isdigit():
          raise IndexError('bad index %s in %s' % (i, '.'.join(name)))
        if int(i) >= l:
          raise IndexError(
            'element %s in %s out of range' % (i, '.'.join(name)))

        e = '"' + i + '":' + args2json(
          name + ('[' + i + ']',), etype, [], x, defarg.get(i, ...))
        json.append(e)

      if kind in ('array') and len(json) < l:
        raise IndexError('%s must have %d elements' % ('.'.join(name), l))

      return '{' + ','.join(json) + '}'

  elif kind in ('optional'):
    # fetch value
    v = kwargs
    if v is ...:
      if len(args): v = args.pop(0)
      elif defarg is not ...: v = defarg
      else: raise TypeError('missing value for %s' % '.'.join(name))

    if v is None: return "null"
    return args2json(name, fmt['type'], [], v, defarg)

  else:
    # fetch value
    v = kwargs
    if v is ...:
      if len(args): v = args.pop(0)
      elif defarg is not ...: v = defarg
      else: raise TypeError('missing value for %s' % '.'.join(name))

    # convert
    try:
      if kind in ('enum'):
        json = '"' + str(v) + '"'
        if not v in fmt['values']:
          if not [k for k in fmt['values'] if k.endswith('::' + str(v))]:
            raise TypeError('should be in ' + ', '.join(fmt['values']))

      elif kind in ('char', 'string'):
        json = '"' + str(v) + '"'
        l = fmt.get('length', len(v))
        if len(v) > l:
          raise IndexError('%s too long (%d max)' % ('.'.join(name), l))

      elif kind in ('double', 'float'):
        json = str(float(v))

      else:
        json = str(int(v))
    except (TypeError, ValueError) as e:
      e.args = ('bad value %s for %s' % (str(v), '.'.join(name)), *e.args)
      raise
    return json


# --- arg2json -------------------------------------------------------------
#
def arg2json(args, kwargs, defarg, fmt):
  """Convert service arguments to a json parameter"""

  # merge flat list and defaults and convert to json
  json = []
  for p, f in fmt.items():
    v = '"' + p + '":' + args2json(
      (p,), f, args, kwargs.pop(p, ...), defarg.get(p, ...))
    json.append(v)

  # check extra arguments
  if args:
    raise TypeError('too many arguments: %s' % args)
  if kwargs:
    p, f = kwargs.popitem()
    raise TypeError('unknown argument %s' % p)

  return '{' + ','.join(json) + '}'


# --- meta2defarg ----------------------------------------------------------
#
def meta2defarg(meta):
  """Extract default argument values from service metadata."""

  defarg = {}
  for p, v in meta.items():
    if 'default' in v:
      defarg[p] = v['default']
    elif 'items' in v:
      for k, v in meta2defarg(v['items']).items():
        defarg.setdefault(p, {})[k] = v

  return defarg


# --- dict2doc -----------------------------------------------------------
#
def dict2doc(indent, name, format, meta):
  """Generate documentation from input metadata"""
  items = {}

  doc = "\n%s%s" % (indent, format['kind'])
  if name is not None:
    doc += " %s" % name
  if 'length' in format:
    doc += "[%d]" % format['length']
  if 'doc' in meta:
    doc += ": %s" % meta['doc']
  if format['kind'] != 'optional':
    if 'default' in meta:
      doc += " (%s)" % meta['default']

  if format['kind'] in ('struct', 'union'):
    for p, f in format['type'].items():
      m = meta.get('items', {}).get(p, {})
      doc += dict2doc(indent + "| ", name + "." + p, f, m)

  elif format['kind'] in ('array', 'sequence'):
    doc += dict2doc(indent + "| ", name + "[]", format['type'], {})

    for i, m in meta.get('items', {}).items():
      doc += dict2doc(indent + "| ", name + "[" + i + "]", format['type'], m)

  elif format['kind'] in ('optional'):
    m = {}
    if 'default' in meta:
      m['default'] = meta['default']

    doc += dict2doc(indent + "| ", "", format['type'], m)

  return doc


# --- meta2doc -----------------------------------------------------------
#
def meta2doc(values, meta):
  """Generate documentation from input metadata"""

  doc = ''
  for p, f in values.items():
    doc += dict2doc("  ", p, f, meta.get(p, {}))

  return doc or '  (empty)'
